(*GLIST
  Fredi Cobo - Software Consultant         fredcobo@yahoo.com
  Sep/04
  Generic list manager*)


MODULE GList;

IMPORT SYSTEM,FIO,Storage,Lib;

TYPE
CompareMask= PROCEDURE(x,y: SYSTEM.PTR): BOOLEAN;
TaskMask*  = PROCEDURE(x: SYSTEM.PTR);

NodePtr    = POINTER TO Node;
Node       = RECORD
               prev,next: NodePtr;
               eob      : BOOLEAN;
               data     : SYSTEM.PTR
             END;

List*      = POINTER TO ListCtl;
ListCtl    = RECORD
               size         : INTEGER;
               first,last   : NodePtr;
               count        : LONGINT;
               p,aux,aux2   : NodePtr;

               Greater,Equal: CompareMask
             END;

PROCEDURE Error(s: ARRAY OF CHAR);
CONST
FLE= "ERR.TXT";

VAR
h : FIO.File;
p0: ARRAY 128 OF CHAR;
BEGIN
  Lib.ParamStr(p0,0);

    IF FIO.Exists(FLE) THEN h:=FIO.Append(FLE)
  ELSE                      h:=FIO.Create(FLE)
   END;

  FIO.WrStr(h,p0);FIO.WrStr(h,": ");FIO.WrStr(h,s);FIO.WrLn(h);

  FIO.Close(h);HALT
END Error;

PROCEDURE GreaterNull*(x,y: SYSTEM.PTR): BOOLEAN;
BEGIN
  Error("GList/GreaterNull/1");RETURN FALSE
END GreaterNull;

PROCEDURE EqualNull*(x,y: SYSTEM.PTR): BOOLEAN;
BEGIN
  Error("GList/EqualNull/1");RETURN FALSE
END EqualNull;

PROCEDURE FirstItem(VAR l   : List;
                        data: ARRAY OF SYSTEM.BYTE);
VAR
p: NodePtr;
BEGIN
  NEW(p);

  l.first:=p;
  l.last :=p;
  l.p    :=p;
  l.count:=1;

  p.prev :=NIL;
  p.next :=NIL;
  p.eob  :=FALSE;

  Storage.ALLOCATE(p.data,l.size);Lib.Move(SYSTEM.ADR(data),p.data,l.size)
END FirstItem;

PROCEDURE LastItem(VAR l   : List;
                       data: ARRAY OF SYSTEM.BYTE);
VAR
p: NodePtr;
BEGIN
  NEW(p);INC(l.count);

  l.last.next:=  p;
       p.prev:=l.last;
       p.next:=  NIL;
       p.eob :=  FALSE;
  l.last     :=  p;
  l.p        :=  p;

  Storage.ALLOCATE(p.data,l.size);Lib.Move(SYSTEM.ADR(data),p.data,l.size)
END LastItem;

PROCEDURE BeforeFirstItem(VAR l   : List;
                              data: ARRAY OF SYSTEM.BYTE);
VAR
p: NodePtr;
BEGIN
  NEW(p);INC(l.count);

        p.prev:=  NIL;
        p.next:=l.first;
        p.eob :=  FALSE;
  l.first.prev:=  p;
  l.first     :=  p;
  l.p         :=  p;

  Storage.ALLOCATE(p.data,l.size);Lib.Move(SYSTEM.ADR(data),p.data,l.size)
END BeforeFirstItem;

PROCEDURE InternalItem(VAR l   : List;
                           data: ARRAY OF SYSTEM.BYTE);
VAR
p,r: NodePtr;
BEGIN
  (*Point to item greater than this*)
  r:=l.first;
  WHILE l.Greater(SYSTEM.ADR(data),r.data) DO
    r:=r.next
  END;

  (*Insert item*)
  NEW(p);INC(l.count);

  p.prev     :=r.prev;
  p.next     :=  r;
  p.eob      :=  FALSE;
  r.prev.next:=  p;
  r.prev     :=  p;
  l.p        :=  p;

  Storage.ALLOCATE(p.data,l.size);Lib.Move(SYSTEM.ADR(data),p.data,l.size)
END InternalItem;

(*----------------------------------------------------------------------------*)

PROCEDURE Create*(VAR l            : List;
                      size         : INTEGER;
                      Greater,Equal: CompareMask);

BEGIN
  NEW(l);

  l.size   :=size;
  l.first  :=NIL;
  l.last   :=NIL;
  l.count  :=0;

  l.Greater:=Greater;
  l.Equal  :=Equal
END Create;

PROCEDURE (l:List) Add*(data: ARRAY OF SYSTEM.BYTE);
BEGIN
    IF l.first=NIL THEN FirstItem(l,data)
  ELSE                   LastItem(l,data)
   END
END Add;

PROCEDURE (l:List) Insert*(data: ARRAY OF SYSTEM.BYTE);
BEGIN
     IF   l.first=NIL                              THEN       FirstItem(l,data)
  ELSIF ~ l.Greater(SYSTEM.ADR(data),l.first.data) THEN BeforeFirstItem(l,data)
  ELSIF   l.Greater(SYSTEM.ADR(data),l.last.data ) THEN        LastItem(l,data)
   ELSE                                                    InternalItem(l,data)
    END
END Insert;

PROCEDURE (l:List) Find*(VAR data: ARRAY OF SYSTEM.BYTE): BOOLEAN;
BEGIN
  l.p:=l.first;
  WHILE l.p#NIL DO
    IF l.Equal(SYSTEM.ADR(data),l.p.data) THEN
       Lib.Move(l.p.data,SYSTEM.ADR(data),l.size);RETURN TRUE
    END;

    l.p:=l.p.next
  END;

  RETURN FALSE
END Find;

PROCEDURE (l:List) Update*(data: ARRAY OF SYSTEM.BYTE);
BEGIN
  IF l.p=NIL THEN
     Error("GList/Update/1")
  END;

  Lib.Move(SYSTEM.ADR(data),l.p.data,l.size)
END Update;

PROCEDURE (l:List) This*(VAR data: ARRAY OF SYSTEM.BYTE);
BEGIN
  IF l.p=NIL THEN
     Error("GList/This/1")
  END;

  Lib.Move(l.p.data,SYSTEM.ADR(data),l.size)
END This;

PROCEDURE (l:List) DeletePtr*;
BEGIN
     IF  (l.p.prev=NIL) & (l.p.next=NIL) THEN
          l.first:=NIL;
          l.last :=NIL
  ELSIF  (l.p.prev=NIL)                  THEN
          l.first      :=l.p.next;
          l.p.next.prev:=NIL
  ELSIF  (l.p.next=NIL)                  THEN
          l.last         :=l.p.prev;
          l.p.prev^.next:=NIL
   ELSE
          l.p.prev^.next:=l.p.next;
          l.p.next^.prev:=l.p.prev
    END;

  DEC(l.count);

  Storage.DEALLOCATE(l.p.data,l.size    );
  Storage.DEALLOCATE(l.p     ,SIZE(Node))
END DeletePtr;

PROCEDURE (l:List) Delete*(data: ARRAY OF SYSTEM.BYTE);
BEGIN
  IF l.Find(data) THEN
     l.DeletePtr
  END
END Delete;

PROCEDURE (l:List) Sort*;
VAR
aux: SYSTEM.ADDRESS;

  PROCEDURE Swap(): BOOLEAN;
  VAR
  q: NodePtr;

    PROCEDURE Add;
    VAR
    p: NodePtr;
    BEGIN
      NEW(p);INC(l.count);

      l.last.next:=  p;
           p.prev:=l.last;
           p.next:=  NIL;
      l.last     :=  p;

      Storage.ALLOCATE(p.data,l.size);Lib.Move(aux,p.data,l.size)
    END Add;

  BEGIN
    l.p:=l.first;
    WHILE l.p.next#NIL DO
      q:=l.p.next;
      WHILE q#NIL DO
        IF l.Greater(l.p.data,q.data) THEN
           Lib.Move(l.p.data,aux,l.size);
           l.DeletePtr;
           Add;

           RETURN TRUE
        END;

        q:=q.next
      END;

      l.p:=l.p.next
    END;

    RETURN FALSE
  END Swap;

BEGIN
  IF l.count<=1 THEN
     RETURN
  END;

  Storage.ALLOCATE(aux,l.size);

  LOOP
    IF ~ Swap() THEN
         Storage.DEALLOCATE(aux,l.size);RETURN
    END
  END
END Sort;

PROCEDURE (l:List) Cycle*(Task: TaskMask);
BEGIN
  l.p:=l.first;
  WHILE l.p#NIL DO
    Task(l.p.data);

    l.p:=l.p.next
  END
END Cycle;

PROCEDURE (l:List) Empty*(): BOOLEAN;
BEGIN
  RETURN l.first=NIL
END Empty;

PROCEDURE (l:List) Count*(): LONGINT;
BEGIN
  RETURN l.count
END Count;

PROCEDURE (l:List) Reset*;
BEGIN
  l.p   :=l.first;
  l.aux :=l.first;
  l.aux2:=l.last
END Reset;

PROCEDURE (l:List) xReset*;
BEGIN
  l.aux :=l.p.next;
  l.aux2:=l.p.prev
END xReset;

PROCEDURE (l:List) Block*(ptr: SYSTEM.ADDRESS);
BEGIN
  Lib.Move(SYSTEM.ADR(ptr),SYSTEM.ADR(l.aux),4);l.p:=NIL
END Block;

PROCEDURE (l:List) First*(VAR data: ARRAY OF SYSTEM.BYTE);
BEGIN
  IF l.first=NIL THEN
     Error("GList/First/1")
  END;

  Lib.Move(l.first.data,SYSTEM.ADR(data),l.size);l.p:=l.first
END First;

PROCEDURE (l:List) Last*(VAR data: ARRAY OF SYSTEM.BYTE);
BEGIN
  IF l.last=NIL THEN
     Error("GList/Last/1")
  END;

  Lib.Move(l.last.data,SYSTEM.ADR(data),l.size);l.p:=l.last
END Last;

PROCEDURE (l:List) Next*(VAR data: ARRAY OF SYSTEM.BYTE): BOOLEAN;
BEGIN
  IF (l.aux=NIL) OR ( (l.p#NIL) & l.p^.eob) THEN
     RETURN FALSE
  END;

  Lib.Move(l.aux.data,SYSTEM.ADR(data),l.size);l.p:=l.aux;l.aux2:=l.aux^.prev;

  l.aux:=l.aux.next;RETURN TRUE
END Next;

PROCEDURE (l:List) Prev*(VAR data: ARRAY OF SYSTEM.BYTE): BOOLEAN;
BEGIN
  IF l.aux2=NIL THEN
     RETURN FALSE
  END;

  Lib.Move(l.aux2.data,SYSTEM.ADR(data),l.size);l.p:=l.aux2;l.aux:=l.aux2^.next;

  l.aux2:=l.aux2.prev;RETURN TRUE
END Prev;

PROCEDURE (l:List) Ptr*(): SYSTEM.ADDRESS;
BEGIN
  RETURN l.p
END Ptr;

PROCEDURE (l:List) EOB*(): BOOLEAN;
BEGIN
  RETURN l.p^.eob
END EOB;

PROCEDURE (l:List) SetEOB*;
BEGIN
  IF l.p^.prev=NIL THEN
     RETURN
  END;

  l.p^.prev^.eob:=TRUE
END SetEOB;

PROCEDURE (l:List) Free*;
VAR
aux: NodePtr;
BEGIN
  WHILE l.first#NIL DO
    aux:=l.first;l.first:=l.first.next;

    Storage.DEALLOCATE(aux.data,l.size    );
    Storage.DEALLOCATE(aux     ,SIZE(Node))
  END;

  l.first:=NIL;
  l.last :=NIL;
  l.count:=0
END Free;

BEGIN
END GList.
