(*XLS
  Fredi Cobo - Software Consultant         fredcobo@yahoo.com
  Mar/05
  Support for Excel 2.1*)

<*+M2EXTENSIONS*>
<*+M2ADDTYPES*>


IMPLEMENTATION MODULE xls;

FROM SYSTEM IMPORT BYTE,ADR,ADDRESS;
FROM Lib    IMPORT Move,AddAddr,ParamStr;
FROM Str    IMPORT Copy,Prepend,Concat,Compare,Length,Count,Caps,ItemS,Slice;
FROM GList  IMPORT List,GreaterNull,EqualNull;
IMPORT FIO,GList;

CONST
EOS                = 0;
BLANK              = 32;

HDR                = 96;               (*header block #1                      *)
HD2                = 404;              (*header block #2                      *)
FON                = 80;               (*default fonts                        *)
FMT                = 29;               (*extra formats                        *)
WPR                = 6;                (*window protect, not used             *)
PRO                = 6;                (*protect worksheet, enable password   *)
XFR                = 8;                (*default XF record                    *)
DCW                = 6;                (*default column width                 *)
DIM                = 12;               (*dimensions                           *)
TAIL               = 55;               (*tail                                 *)

     ROW_OFS       = 20;               (*location for row block offset        *)
WPROTECT_OFS       = 12;               (*location for window protection offset*)

STRLEN             = 255;
RECLEN             = 12 + STRLEN;      (*label record: header + string        *)

     INDEX_FIRSTROW= 16;
     INDEX_LASTROW = 18;

DIMENSIONS_FIRSTROW= 4;
DIMENSIONS_LASTROW = 6;
DIMENSIONS_FIRSTCOL= 8;
DIMENSIONS_LASTCOL = 10;

FORMAT_COUNT       = 15H;
  FONT_COUNT       = 4H;

POINT              = 14H;    (*1 point font height, default is 10 points: 0C8H*)
CHARWIDTH          = 120H;   (*1 char width,        default is 10 chars       *)

DATE               = 101;    (*date id format: dd-mmm-yyyy                    *)

KEYS               = 2;

TYPE
STRING   = ARRAY[0..128] OF CHAR;

FontsNode= RECORD
             font: Font;
             name: STRING
           END;

PicNode  = RECORD
             ndx      : CARDINAL;   (*1 based, 0 is the default XF record*)
             color    : Color;
             font     : Font;
             points   : CARDINAL;
             style    : BYTE;
             align    : BYTE;
             decimales: CARDINAL
           END;

CellNode = RECORD
             row,col,i: CARDINAL;
             record   : ARRAY[0..RECLEN-1] OF BYTE
           END;

WidNode  = RECORD
             first,last: BYTE;
             width     : CARDINAL
           END;

RowNode  = RECORD
             n               : CARDINAL;
             firstCol,lastCol: CARDINAL;
           END;

VAR
Fonts,Pic,Cell,Wid,Row: List;

file                  : STRING;
firstRow,lastRow      : CARDINAL;
firstCol,lastCol      : CARDINAL;
password              : CARDINAL;

hdr                   : ARRAY[0..HDR -1] OF BYTE;
hd2                   : ARRAY[0..HD2 -1] OF BYTE;
fon                   : ARRAY[0..FON -1] OF BYTE;
fmt                   : ARRAY[0..FMT -1] OF BYTE;
wpr                   : ARRAY[0..WPR -1] OF BYTE;
pro                   : ARRAY[0..PRO -1] OF BYTE;
xfr                   : ARRAY[0..XFR -1] OF BYTE;
dcw                   : ARRAY[0..DCW -1] OF BYTE;
dim                   : ARRAY[0..DIM -1] OF BYTE;
tail                  : ARRAY[0..TAIL-1] OF BYTE;

key                   : ARRAY[0..KEYS] OF CARDINAL;

cell                  : CellNode; (*globals*)
curr                  : PicNode;
_c                    : CARDINAL;

PROCEDURE EqualFonts(pdata,pnode: ADDRESS): BOOLEAN;
VAR
data,node: FontsNode;
BEGIN
  Move(pdata,ADR(data),SIZE(FontsNode));
  Move(pnode,ADR(node),SIZE(FontsNode));

  RETURN data.font=node.font
END EqualFonts;

PROCEDURE EqualPic(pdata,pnode: ADDRESS): BOOLEAN;
VAR
data,node: PicNode;
BEGIN
  Move(pdata,ADR(data),SIZE(PicNode));
  Move(pnode,ADR(node),SIZE(PicNode));

  RETURN     (data.color    =node.color    )
         AND (data.font     =node.font     )
         AND (data.points   =node.points   )
         AND (data.style    =node.style    )
         AND (data.align    =node.align    )
         AND (data.decimales=node.decimales)
END EqualPic;

PROCEDURE GreaterCell(pdata,pnode: ADDRESS): BOOLEAN;
VAR
data,node: CellNode;
BEGIN
  Move(pdata,ADR(data),SIZE(CellNode));
  Move(pnode,ADR(node),SIZE(CellNode));

     IF data.row>node.row THEN RETURN TRUE
  ELSIF data.row<node.row THEN RETURN FALSE
   ELSE
        RETURN data.col>=node.col
    END
END GreaterCell;

PROCEDURE GreaterRow(pdata,pnode: ADDRESS): BOOLEAN;
VAR
data,node: RowNode;
BEGIN
  Move(pdata,ADR(data),SIZE(RowNode));
  Move(pnode,ADR(node),SIZE(RowNode));

  RETURN data.n>=node.n
END GreaterRow;

PROCEDURE EqualRow(pdata,pnode: ADDRESS): BOOLEAN;
VAR
data,node: RowNode;
BEGIN
  Move(pdata,ADR(data),SIZE(RowNode));
  Move(pnode,ADR(node),SIZE(RowNode));

  RETURN data.n=node.n
END EqualRow;

(*----------------------------------------------------------------------------*)

(*Error log*)
PROCEDURE Error(s: ARRAY OF CHAR);
CONST
FLE= "ERR.TXT";

VAR
h : FIO.File;
p0: STRING;
BEGIN
  ParamStr(p0,0);

    IF FIO.Exists(FLE) THEN h:=FIO.Append(FLE)
  ELSE                      h:=FIO.Create(FLE)
   END;

  FIO.WrStr(h,p0);FIO.WrStr(h,": ");FIO.WrStr(h,s);FIO.WrLn(h);

  FIO.Close(h);HALT
END Error;

(*Trim chars*)
PROCEDURE _Trim(VAR txt: ARRAY OF CHAR);
VAR
k,w,l,r,j: CARDINAL;
ch       : CHAR;
filled   : BOOLEAN;
BEGIN
     IF Compare(txt,"" )=0 THEN Copy(txt,"");_c:=BLANK;RETURN
  ELSIF Compare(txt," ")=0 THEN Copy(txt,"");_c:=BLANK;RETURN
  ELSIF HIGH(txt)=0        THEN Copy(txt,"");_c:=BLANK;RETURN
    END;

  txt[HIGH(txt)]:=CHR(EOS);ch:=CHR(_c);_c:=BLANK;

  w:=Length(txt);
  IF w=0 THEN
     txt[0]:=CHR(EOS);RETURN
  END;

  filled:=TRUE;
  FOR k:=0 TO w-1 DO
    filled:=filled AND (txt[k]=ch)
  END;
  IF filled THEN
     txt[0]:=CHR(EOS);RETURN
  END;

  (*left pointer*)
  l:=0;
  LOOP
    IF txt[l]#ch THEN EXIT END;
    INC(l)
  END;

  (*right pointer*)
  r:=w-1;
  LOOP
    IF txt[r]#ch THEN EXIT END;
    DEC(r)
  END;

  (*copy string*)
  j:=0;
  FOR k:=l TO r DO
    txt[j]:=txt[k];INC(j)
  END;
  txt[j]:=CHR(EOS)
END _Trim;

(*----------------------------------------------------------------------------*)

PROCEDURE Init;

  PROCEDURE Fuentes;
  VAR
  fonts: FontsNode;
  BEGIN
    fonts.font:=arial  ;fonts.name:="Arial"          ;Fonts^.Add(fonts);
    fonts.font:=courier;fonts.name:="Courier New"    ;Fonts^.Add(fonts);
    fonts.font:=times  ;fonts.name:="Times New Roman";Fonts^.Add(fonts);
    fonts.font:=lucida ;fonts.name:="Lucida Sans"    ;Fonts^.Add(fonts);
    fonts.font:=tahoma ;fonts.name:="Tahoma"         ;Fonts^.Add(fonts);
  END Fuentes;

  PROCEDURE Keys;
  BEGIN
    key[0]:=0;                                                   (*no password*)

    key[1]:=0957DH;                                              (*onegroup   *)
    key[2]:=0DE69H;                                              (*xantia     *)
  END Keys;

BEGIN
  GList.Create(Fonts,SIZE(FontsNode),GreaterNull,EqualFonts);
  GList.Create(Pic  ,SIZE(  PicNode),GreaterNull,EqualPic  );
  GList.Create(Cell ,SIZE( CellNode),GreaterCell,EqualNull );
  GList.Create(Wid  ,SIZE(  WidNode),GreaterNull,EqualNull );
  GList.Create(Row  ,SIZE(  RowNode),GreaterRow ,EqualRow  );

  Fuentes;Keys
END Init;

PROCEDURE Min(n1,n2: CARDINAL): CARDINAL;
BEGIN
    IF n1<n2 THEN RETURN n1
  ELSE            RETURN n2
   END
END Min;

PROCEDURE Max(n1,n2: CARDINAL): CARDINAL;
BEGIN
    IF n1>n2 THEN RETURN n1
  ELSE            RETURN n2
   END
END Max;

PROCEDURE OpCode(opcode: CARDINAL;
                 length: CARDINAL);
BEGIN
  Move(ADR(opcode),        ADR(cell.record)   ,2);
  Move(ADR(length),AddAddr(ADR(cell.record),2),2);

  cell.i:=4
END OpCode;

PROCEDURE Byte(n: BYTE);
BEGIN
  Move(ADR(n),AddAddr(ADR(cell.record),cell.i),1);INC(cell.i)
END Byte;

PROCEDURE Card(n: CARDINAL);
BEGIN
  Move(ADR(n),AddAddr(ADR(cell.record),cell.i),2);INC(cell.i,2)
END Card;

PROCEDURE IEEE(n: LONGREAL);
BEGIN
  Move(ADR(n),AddAddr(ADR(cell.record),cell.i),8);INC(cell.i,8)
END IEEE;

PROCEDURE InsertCell(r,c: CARDINAL);
VAR
row: RowNode;
BEGIN
  (*globals*)
  firstRow:=Min(firstRow,r);lastRow:=Max(lastRow,r);
  firstCol:=Min(firstCol,c);lastCol:=Max(lastCol,c);

  (*cell list*)
  cell.row:=r;cell.col:=c;Cell^.Insert(cell);

  (*row list*)
  row.n:=r;
  IF NOT Row^.Find(row) THEN
     row.firstCol:=c;
     row.lastCol :=c;

     Row^.Insert(row)
  ELSE
     row.firstCol:=Min(row.firstCol,c);
     row.lastCol :=Max(row.lastCol ,c);

     Row^.Update(row)
  END
END InsertCell;

(*----------------------------------------------------------------------------*)

(*New worksheet*)
PROCEDURE Create(s: ARRAY OF CHAR;
                 p: CARDINAL);

  PROCEDURE SetupFile;
  VAR
  n  : CARDINAL;
  ext: STRING;
  BEGIN
    _Trim(s);Caps(s);

    n:=Count(s,".");ItemS(ext,s,".",1);Prepend(ext,".");

       IF n=0                THEN Concat(file,s,EXT)
    ELSIF n>1                THEN Error("xls/Create/SetupFile/1")
    ELSIF Compare(ext,EXT)#0 THEN Error("xls/Create/SetupFile/2")
     ELSE                         Copy(file,s)
      END
  END SetupFile;

BEGIN
  SetupFile;password:=p;

  firstRow:=MAX(CARDINAL);lastRow:=0;
  firstCol:=MAX(CARDINAL);lastCol:=0;

   Row^.Free;
  Cell^.Free;
   Pic^.Free;
   Wid^.Free
END Create;

(*Close file*)
PROCEDURE Close(): BOOLEAN;
VAR
h  : FIO.File;
ofs: LONGCARD;

  PROCEDURE LongCreate(): BOOLEAN; (*Don't use on NETWARE drives, support only short names 8.3*)

    PROCEDURE Delete(): BOOLEAN;
    VAR
    d: FIO.DirEntry;
    BEGIN
      FIO.IOcheck:=FALSE;

         IF NOT FIO.ReadFirstEntry(file,FIO.FileAttr{},d) THEN FIO.IOcheck:=TRUE;RETURN TRUE
      ELSIF FIO.readonly IN d.attr                        THEN FIO.IOcheck:=TRUE;RETURN FALSE
        END;

      FIO.Erase(file);

      FIO.IOcheck:=TRUE;RETURN FIO.IOresult()=0
    END Delete;

  BEGIN
    IF NOT Delete() THEN
       RETURN FALSE
    END;

    h:=FIO.Create(file);ofs:=0;

    RETURN FIO.IOresult()=0
  END LongCreate;

  PROCEDURE WrBin(VAR data: ARRAY OF BYTE;
                      n   : CARDINAL);
  BEGIN
    FIO.WrBin(h,data,n);INC(ofs,n)
  END WrBin;

  PROCEDURE Patch(pos: LONGCARD);
  VAR
  data: ARRAY[0..2] OF BYTE;
  BEGIN
    Move(ADR(ofs),ADR(data),2);

    FIO.Seek(h,pos);FIO.WrBin(h,data,2);

    FIO.Seek(h,ofs)
  END Patch;

  PROCEDURE WrHeader;

    PROCEDURE Patch(ofs,n: CARDINAL);
    BEGIN
      Move(ADR(n),AddAddr(ADR(hdr),ofs),2)
    END Patch;

  BEGIN
    Patch(INDEX_FIRSTROW,firstRow    );
    Patch(INDEX_LASTROW , lastRow + 1);

    WrBin(hdr,HDR)
  END WrHeader;

  PROCEDURE WrHeader2;
  BEGIN
    WrBin(hd2,HD2)
  END WrHeader2;

  PROCEDURE WrDefaultFonts;
  BEGIN
    WrBin(fon,FON)
  END WrDefaultFonts;

  PROCEDURE WrUserFonts;
  VAR
  fonts: FontsNode;
  pic  : PicNode;
  k    : CARDINAL;
  BEGIN
    Pic^.Reset;
    WHILE Pic^.Next(pic) DO
      fonts.font:=pic.font;
      IF NOT Fonts^.Find(fonts) THEN
         Error("xls/Close/WrUserFonts/1")
      END;

      (*FONT*)
      OpCode(31H,5 + Length(fonts.name));
      Card(pic.points*POINT);                               (*font height     *)
      Byte(pic.style);                                      (*font attribute  *)
      Byte(0H);                                             (*reserved        *)
      Byte(BYTE(Length(fonts.name)));                       (*font name length*)

      FOR k:=0 TO Length(fonts.name)-1 DO                   (*font name       *)
        Byte(BYTE(fonts.name[k]))
      END;

      WrBin(cell.record,cell.i);

      (*EFONT*)
      OpCode(45H,2H);
      Card(ORD(pic.color));

      WrBin(cell.record,cell.i)
    END
  END WrUserFonts;

  PROCEDURE WrUserFormats;
  VAR
  pic: PicNode;

    PROCEDURE Format(decimales: CARDINAL);
    VAR
    s,ceros: STRING;
    k      : CARDINAL;
    BEGIN
      CASE decimales OF
         0   : s:="#,##0"
        |DATE: s:="d\-mmm\-yyyy"
      ELSE
               Slice(ceros,"0000000000000000",0,decimales);Concat(s,"#,##0.",ceros)
      END;

      OpCode(1EH,Length(s) + 1);
      Byte(BYTE(Length(s)));

      FOR k:=0 TO Length(s)-1 DO
        Byte(BYTE(s[k]))
      END;

      WrBin(cell.record,cell.i)
    END Format;

  BEGIN
    Pic^.Reset;
    WHILE Pic^.Next(pic) DO
      Format(pic.decimales)
    END
  END WrUserFormats;

  PROCEDURE WrExtraFormats;
  BEGIN
    WrBin(fmt,FMT)
  END WrExtraFormats;

  PROCEDURE WrProtect;
  BEGIN
    IF password#0 THEN
       WrBin(pro,PRO)
    END
  END WrProtect;

  PROCEDURE WrWindowProtect;
  BEGIN
    Patch(WPROTECT_OFS);WrBin(wpr,WPR)
  END WrWindowProtect;

  PROCEDURE WrPassword;
  BEGIN
       IF password=0    THEN RETURN
    ELSIF password>KEYS THEN Error("xls/Close/WrPassword/1")
      END;

    OpCode(13H,2H);
    Card(key[password]);

    WrBin(cell.record,cell.i)
  END WrPassword;

  PROCEDURE WrDefaultXF;
  BEGIN
    WrBin(xfr,XFR)
  END WrDefaultXF;

  PROCEDURE WrUserXF;
  VAR
  pic: PicNode;

    PROCEDURE Format(pic: PicNode);

      PROCEDURE FormatIndex;
      BEGIN
        (*
        7654 3210
        *#xx xxxx
        |||
        ||+-----------  index in FORMAT record list
        ++------------  cell is locked
        +-------------  cell is hidden
        *)

        Byte(BYTE( 40H                               (*0100 0000 default value*)
                  +FORMAT_COUNT + pic.ndx - 1))
      END FormatIndex;

    BEGIN
      OpCode(43H,4H);

      Byte(BYTE(FONT_COUNT + pic.ndx));      (*index to FONT record           *)
      Byte(0H);                              (*gridlines codes for Kanji EXCEL*)

      FormatIndex;

      Byte(pic.align);                       (*alignment flags                *)

      WrBin(cell.record,cell.i)
    END Format;

  BEGIN
    Pic^.Reset;
    WHILE Pic^.Next(pic) DO
      Format(pic)
    END
  END WrUserXF;

  PROCEDURE WrDefaultColumnWidth;
  BEGIN
    WrBin(dcw,DCW)
  END WrDefaultColumnWidth;

  PROCEDURE WrUserColumnWidth;
  VAR
  wid: WidNode;
  BEGIN
    Wid^.Reset;
    WHILE Wid^.Next(wid) DO
      OpCode(24H,4H);
      Byte(wid.first);
      Byte(wid.last );
      Card(wid.width);

      WrBin(cell.record,cell.i)
    END
  END WrUserColumnWidth;

  PROCEDURE WrDimensions;

    PROCEDURE Patch(ofs,n: CARDINAL);
    BEGIN
      Move(ADR(n),AddAddr(ADR(dim),ofs),2)
    END Patch;

  BEGIN
    Patch(DIMENSIONS_FIRSTROW,firstRow    );
    Patch(DIMENSIONS_LASTROW , lastRow + 1);
    Patch(DIMENSIONS_FIRSTCOL,firstCol    );
    Patch(DIMENSIONS_LASTCOL , lastCol + 1);

    WrBin(dim,DIM)
  END WrDimensions;

  PROCEDURE WrTail;
  BEGIN
    WrBin(tail,TAIL)
  END WrTail;

  PROCEDURE WrRows;
  VAR
  row: RowNode;
  BEGIN
    Patch(ROW_OFS);

    Row^.Reset;
    WHILE Row^.Next(row) DO
      OpCode(8H,0DH);

      Card(row.n          );
      Card(row.firstCol   );
      Card(row.lastCol + 1);
      Card(    80FFH      );            (*row height                          *)
      Card(       0H      );            (*reserved                            *)
      Byte(       0H      );            (*standard cell attributes            *)
      Card(       0H      );            (*relative file offset to cell records*)

      WrBin(cell.record,cell.i)
    END
  END WrRows;

  PROCEDURE WrCells;
  VAR
  cell: CellNode;
  BEGIN
    Cell^.Reset;
    WHILE Cell^.Next(cell)  DO
      WrBin(cell.record,cell.i)
    END
  END WrCells;

  PROCEDURE Free;
  BEGIN
     Row^.Free;
    Cell^.Free;
     Pic^.Free;
     Wid^.Free
  END Free;

BEGIN
  IF NOT LongCreate() THEN
     RETURN FALSE
  END;

  WrHeader;
  WrDefaultFonts;
  WrUserFonts;
  WrHeader2;
  WrUserFormats;
  WrExtraFormats;
  WrProtect;
  WrWindowProtect;
  WrPassword;
  WrDefaultXF;
  WrUserXF;
  WrDefaultColumnWidth;
  WrUserColumnWidth;
  WrDimensions;
  WrRows;
  WrCells;
  WrTail;

  FIO.Close(h);Free;RETURN TRUE
END Close;

(*Cells*)
PROCEDURE Label(r,c   : CARDINAL;
                label : ARRAY OF CHAR;
                color : Color;
                font  : Font;
                points: CARDINAL;
                style : BYTE;
                align : BYTE);
CONST
decimales= 0;

VAR
pic: PicNode;
l  : CARDINAL;
k  : INTEGER;

  PROCEDURE CurrentSettings;
  BEGIN
    curr.color    :=color;
    curr.font     :=font;
    curr.points   :=points;
    curr.style    :=style;
    curr.align    :=align;
    curr.decimales:=decimales
  END CurrentSettings;

  PROCEDURE Attribute0;
  BEGIN
    (*
    7654 3210
    *#xx xxxx
    |||
    ||+-----------  index in XF record list
    |+------------  cell is locked
    +-------------  cell is hidden
    *)

    pic.color    :=color;
    pic.font     :=font;
    pic.points   :=points;
    pic.style    :=style;
    pic.align    :=align;
    pic.decimales:=decimales;

    IF NOT Pic^.Find(pic) THEN
       pic.ndx:=Pic^.Count() + 1;Pic^.Add(pic)
    END;

    Byte(BYTE( 40H                                   (*0100 0000 default value*)
              +pic.ndx))
  END Attribute0;

  PROCEDURE Attribute1;
  BEGIN
    (*
    7654 3210
    **xx xxxx
    | |
    | +-----------  index in FORMAT record list
    +-------------  index in FONT   record list
    *)

    Byte(BYTE( 0H                                    (*0000 0000 default value*)
              +FORMAT_COUNT + pic.ndx - 1))
  END Attribute1;

  PROCEDURE Attribute2;
  BEGIN
    Byte(pic.align)                                          (*alignment flags*)
  END Attribute2;

BEGIN
  CurrentSettings;

  _Trim(label);l:=Length(label);

  IF Length(label)>STRLEN THEN
     Error("xls/Label/1")
  END;

  OpCode(4H,8+Length(label));

  Card(r);                                                  (*row             *)
  Card(c);                                                  (*column          *)

  Attribute0;                                               (*attributes      *)
  Attribute1;
  Attribute2;

  Byte(BYTE(l));                                            (*length of string*)

  FOR k:=0 TO INTEGER(l)-1 DO                               (*string          *)
    Byte(BYTE(label[k]))
  END;

  InsertCell(r,c)
END Label;

PROCEDURE WrLabel(r,c : CARDINAL;
                 label: ARRAY OF CHAR);
BEGIN
  Label(r,c,label,curr.color,curr.font,curr.points,curr.style,curr.align)
END WrLabel;

PROCEDURE Cardinal(r,c   : CARDINAL;
                   n     : CARDINAL;
                   color : Color;
                   font  : Font;
                   points: CARDINAL;
                   style : BYTE;
                   align : BYTE);
CONST
decimales= 0;

VAR
pic: PicNode;

  PROCEDURE CurrentSettings;
  BEGIN
    curr.color    :=color;
    curr.font     :=font;
    curr.points   :=points;
    curr.style    :=style;
    curr.align    :=align;
    curr.decimales:=decimales
  END CurrentSettings;

  PROCEDURE Attribute0;
  BEGIN
    (*
    7654 3210
    *#xx xxxx
    |||
    ||+-----------  index in XF record list
    |+------------  cell is locked
    +-------------  cell is hidden
    *)

    pic.color    :=color;
    pic.font     :=font;
    pic.points   :=points;
    pic.style    :=style;
    pic.align    :=align;
    pic.decimales:=decimales;

    IF NOT Pic^.Find(pic) THEN
       pic.ndx:=Pic^.Count() + 1;Pic^.Add(pic)
    END;

    Byte(BYTE( 40H                                   (*0100 0000 default value*)
              +pic.ndx))
  END Attribute0;

  PROCEDURE Attribute1;
  BEGIN
    (*
    7654 3210
    **xx xxxx
    | |
    | +-----------  index in FORMAT record list
    +-------------  index in FONT   record list
    *)

    Byte(BYTE( 0H                                    (*0000 0000 default value*)
              +FORMAT_COUNT + pic.ndx - 1))
  END Attribute1;

  PROCEDURE Attribute2;
  BEGIN
    Byte(pic.align)                                          (*alignment flags*)
  END Attribute2;

BEGIN
  CurrentSettings;

  OpCode(2H,9H);

  Card(r);                                                       (*row        *)
  Card(c);                                                       (*column     *)

  Attribute0;                                                    (*attributes *)
  Attribute1;
  Attribute2;

  Card(n);                                                       (*julian date*)

  InsertCell(r,c)
END Cardinal;

PROCEDURE WrCardinal(r,c: CARDINAL;
                     n  : CARDINAL);
BEGIN
  Cardinal(r,c,n,curr.color,curr.font,curr.points,curr.style,curr.align)
END WrCardinal;

PROCEDURE Float(r,c      : CARDINAL;
                n        : LONGREAL;
                color    : Color;
                font     : Font;
                points   : CARDINAL;
                style    : BYTE;
                align    : BYTE;
                decimales: CARDINAL);
VAR
pic: PicNode;

  PROCEDURE CurrentSettings;
  BEGIN
    curr.color    :=color;
    curr.font     :=font;
    curr.points   :=points;
    curr.style    :=style;
    curr.align    :=align;
    curr.decimales:=decimales
  END CurrentSettings;

  PROCEDURE Attribute0;
  BEGIN
    (*
    7654 3210
    *#xx xxxx
    |||
    ||+-----------  index in XF record list
    |+------------  cell is locked
    +-------------  cell is hidden
    *)

    pic.color    :=color;
    pic.font     :=font;
    pic.points   :=points;
    pic.style    :=style;
    pic.align    :=align;
    pic.decimales:=decimales;

    IF NOT Pic^.Find(pic) THEN
       pic.ndx:=Pic^.Count() + 1;Pic^.Add(pic)
    END;

    Byte(BYTE( 40H                                   (*0100 0000 default value*)
              +pic.ndx))
  END Attribute0;

  PROCEDURE Attribute1;
  BEGIN
    (*
    7654 3210
    **xx xxxx
    | |
    | +-----------  index in FORMAT record list
    +-------------  index in FONT   record list
    *)

    Byte(BYTE( 0H                                    (*0000 0000 default value*)
              +FORMAT_COUNT + pic.ndx - 1))
  END Attribute1;

  PROCEDURE Attribute2;
  BEGIN
    Byte(pic.align)                                          (*alignment flags*)
  END Attribute2;

BEGIN
  CurrentSettings;

  OpCode(3H,0FH);

  Card(r);                                                  (*row             *)
  Card(c);                                                  (*column          *)

  Attribute0;                                               (*attributes      *)
  Attribute1;
  Attribute2;

  IEEE(n);

  InsertCell(r,c)
END Float;

PROCEDURE WrFloat(r,c: CARDINAL;
                  n  : LONGREAL);
BEGIN
  Float(r,c,n,curr.color,curr.font,curr.points,curr.style,curr.align,curr.decimales)
END WrFloat;

PROCEDURE Date(r,c   : CARDINAL;
               n     : CARDINAL;
               color : Color;
               font  : Font;
               points: CARDINAL;
               style : BYTE;
               align : BYTE);
CONST
decimales= DATE;

VAR
pic: PicNode;

  PROCEDURE CurrentSettings;
  BEGIN
    curr.color    :=color;
    curr.font     :=font;
    curr.points   :=points;
    curr.style    :=style;
    curr.align    :=align;
    curr.decimales:=decimales
  END CurrentSettings;

  PROCEDURE Attribute0;
  BEGIN
    (*
    7654 3210
    *#xx xxxx
    |||
    ||+-----------  index in XF record list
    |+------------  cell is locked
    +-------------  cell is hidden
    *)

    pic.color    :=color;
    pic.font     :=font;
    pic.points   :=points;
    pic.style    :=style;
    pic.align    :=align;
    pic.decimales:=decimales;

    IF NOT Pic^.Find(pic) THEN
       pic.ndx:=Pic^.Count() + 1;Pic^.Add(pic)
    END;

    Byte(BYTE( 40H                                   (*0100 0000 default value*)
              +pic.ndx))
  END Attribute0;

  PROCEDURE Attribute1;
  BEGIN
    (*
    7654 3210
    **xx xxxx
    | |
    | +-----------  index in FORMAT record list
    +-------------  index in FONT   record list
    *)

    Byte(BYTE( 0H                               (*0000 0000 default value*)
              +FORMAT_COUNT + pic.ndx - 1))
  END Attribute1;

  PROCEDURE Attribute2;
  BEGIN
    Byte(pic.align)                                          (*alignment flags*)
  END Attribute2;

BEGIN
  CurrentSettings;

  OpCode(2H,9H);

  Card(r);                                                       (*row        *)
  Card(c);                                                       (*column     *)

  Attribute0;                                                    (*attributes *)
  Attribute1;
  Attribute2;

  Card(n);                                                       (*julian date*)

  InsertCell(r,c)
END Date;

PROCEDURE WrDate(r,c: CARDINAL;
                 n  : CARDINAL);
BEGIN
  Date(r,c,n,curr.color,curr.font,curr.points,curr.style,curr.align)
END WrDate;

PROCEDURE Width(first,last,chars: CARDINAL);
VAR
wid: WidNode;
BEGIN
     IF first>255   THEN Error("xls/Width/1")
  ELSIF last >255   THEN Error("xls/Width/2")
  ELSIF last <first THEN Error("xls/Width/3")
    END;

  wid.first:=BYTE(first);
  wid.last :=BYTE(last );
  wid.width:=chars*CHARWIDTH;

  Wid^.Add(wid)
END Width;

BEGIN
  Init;_c:=BLANK;

  hdr[  0]:=  9H;hdr[  1]:=  0H;hdr[  2]:=  4H;hdr[  3]:=  0H;hdr[  4]:=  7H;hdr[  5]:=  0H;hdr[  6]:= 10H;hdr[  7]:=  0H;hdr[  8]:= 0BH;hdr[  9]:=  2H;
  hdr[ 10]:= 0CH;hdr[ 11]:=  0H;hdr[ 12]:=  0H;hdr[ 13]:=  0H;hdr[ 14]:=  0H;hdr[ 15]:=  0H;hdr[ 16]:=  0H;hdr[ 17]:=  0H;hdr[ 18]:=  0H;hdr[ 19]:=  0H;
  hdr[ 20]:=  0H;hdr[ 21]:=  0H;hdr[ 22]:=  0H;hdr[ 23]:=  0H;hdr[ 24]:= 42H;hdr[ 25]:=  0H;hdr[ 26]:=  2H;hdr[ 27]:=  0H;hdr[ 28]:=  1H;hdr[ 29]:= 80H;
  hdr[ 30]:= 0DH;hdr[ 31]:=  0H;hdr[ 32]:=  2H;hdr[ 33]:=  0H;hdr[ 34]:=  1H;hdr[ 35]:=  0H;hdr[ 36]:= 0CH;hdr[ 37]:=  0H;hdr[ 38]:=  2H;hdr[ 39]:=  0H;
  hdr[ 40]:= 64H;hdr[ 41]:=  0H;hdr[ 42]:= 0FH;hdr[ 43]:=  0H;hdr[ 44]:=  2H;hdr[ 45]:=  0H;hdr[ 46]:=  1H;hdr[ 47]:=  0H;hdr[ 48]:= 11H;hdr[ 49]:=  0H;
  hdr[ 50]:=  2H;hdr[ 51]:=  0H;hdr[ 52]:=  0H;hdr[ 53]:=  0H;hdr[ 54]:= 10H;hdr[ 55]:=  0H;hdr[ 56]:=  8H;hdr[ 57]:=  0H;hdr[ 58]:=0FCH;hdr[ 59]:=0A9H;
  hdr[ 60]:=0F1H;hdr[ 61]:=0D2H;hdr[ 62]:= 4DH;hdr[ 63]:= 62H;hdr[ 64]:= 50H;hdr[ 65]:= 3FH;hdr[ 66]:= 0EH;hdr[ 67]:=  0H;hdr[ 68]:=  2H;hdr[ 69]:=  0H;
  hdr[ 70]:=  1H;hdr[ 71]:=  0H;hdr[ 72]:= 22H;hdr[ 73]:=  0H;hdr[ 74]:=  2H;hdr[ 75]:=  0H;hdr[ 76]:=  0H;hdr[ 77]:=  0H;hdr[ 78]:= 2AH;hdr[ 79]:=  0H;
  hdr[ 80]:=  2H;hdr[ 81]:=  0H;hdr[ 82]:=  0H;hdr[ 83]:=  0H;hdr[ 84]:= 2BH;hdr[ 85]:=  0H;hdr[ 86]:=  2H;hdr[ 87]:=  0H;hdr[ 88]:=  0H;hdr[ 89]:=  0H;
  hdr[ 90]:= 25H;hdr[ 91]:=  0H;hdr[ 92]:=  2H;hdr[ 93]:=  0H;hdr[ 94]:=0FFH;hdr[ 95]:= 80H;

  fon[  0]:= 31H;fon[  1]:=  0H;fon[  2]:= 0AH;fon[  3]:=  0H;fon[  4]:=0C8H;fon[  5]:=  0H;fon[  6]:=  0H;fon[  7]:=  0H;fon[  8]:=  5H;fon[  9]:= 41H;
  fon[ 10]:= 72H;fon[ 11]:= 69H;fon[ 12]:= 61H;fon[ 13]:= 6CH;fon[ 14]:= 45H;fon[ 15]:=  0H;fon[ 16]:=  2H;fon[ 17]:=  0H;fon[ 18]:=0FFH;fon[ 19]:= 7FH;
  fon[ 20]:= 31H;fon[ 21]:=  0H;fon[ 22]:= 0AH;fon[ 23]:=  0H;fon[ 24]:=0C8H;fon[ 25]:=  0H;fon[ 26]:=  0H;fon[ 27]:=  0H;fon[ 28]:=  5H;fon[ 29]:= 41H;
  fon[ 30]:= 72H;fon[ 31]:= 69H;fon[ 32]:= 61H;fon[ 33]:= 6CH;fon[ 34]:= 45H;fon[ 35]:=  0H;fon[ 36]:=  2H;fon[ 37]:=  0H;fon[ 38]:=0FFH;fon[ 39]:= 7FH;
  fon[ 40]:= 31H;fon[ 41]:=  0H;fon[ 42]:= 0AH;fon[ 43]:=  0H;fon[ 44]:=0C8H;fon[ 45]:=  0H;fon[ 46]:=  0H;fon[ 47]:=  0H;fon[ 48]:=  5H;fon[ 49]:= 41H;
  fon[ 50]:= 72H;fon[ 51]:= 69H;fon[ 52]:= 61H;fon[ 53]:= 6CH;fon[ 54]:= 45H;fon[ 55]:=  0H;fon[ 56]:=  2H;fon[ 57]:=  0H;fon[ 58]:=0FFH;fon[ 59]:= 7FH;
  fon[ 60]:= 31H;fon[ 61]:=  0H;fon[ 62]:= 0AH;fon[ 63]:=  0H;fon[ 64]:=0C8H;fon[ 65]:=  0H;fon[ 66]:=  0H;fon[ 67]:=  0H;fon[ 68]:=  5H;fon[ 69]:= 41H;
  fon[ 70]:= 72H;fon[ 71]:= 69H;fon[ 72]:= 61H;fon[ 73]:= 6CH;fon[ 74]:= 45H;fon[ 75]:=  0H;fon[ 76]:=  2H;fon[ 77]:=  0H;fon[ 78]:=0FFH;fon[ 79]:= 7FH;

  hd2[  0]:= 14H;hd2[  1]:=  0H;hd2[  2]:=  0H;hd2[  3]:=  0H;hd2[  4]:= 15H;hd2[  5]:=  0H;hd2[  6]:=  0H;hd2[  7]:=  0H;hd2[  8]:= 26H;hd2[  9]:=  0H;
  hd2[ 10]:=  8H;hd2[ 11]:=  0H;hd2[ 12]:=  0H;hd2[ 13]:=  0H;hd2[ 14]:=  0H;hd2[ 15]:=  0H;hd2[ 16]:=  0H;hd2[ 17]:=  0H;hd2[ 18]:=  0H;hd2[ 19]:=  0H;
  hd2[ 20]:= 27H;hd2[ 21]:=  0H;hd2[ 22]:=  8H;hd2[ 23]:=  0H;hd2[ 24]:=  0H;hd2[ 25]:=  0H;hd2[ 26]:=  0H;hd2[ 27]:=  0H;hd2[ 28]:=  0H;hd2[ 29]:=  0H;
  hd2[ 30]:=  0H;hd2[ 31]:=  0H;hd2[ 32]:= 28H;hd2[ 33]:=  0H;hd2[ 34]:=  8H;hd2[ 35]:=  0H;hd2[ 36]:=  0H;hd2[ 37]:=  0H;hd2[ 38]:=  0H;hd2[ 39]:=  0H;
  hd2[ 40]:=  0H;hd2[ 41]:=  0H;hd2[ 42]:=  0H;hd2[ 43]:=  0H;hd2[ 44]:= 29H;hd2[ 45]:=  0H;hd2[ 46]:=  8H;hd2[ 47]:=  0H;hd2[ 48]:=  0H;hd2[ 49]:=  0H;
  hd2[ 50]:=  0H;hd2[ 51]:=  0H;hd2[ 52]:=  0H;hd2[ 53]:=  0H;hd2[ 54]:=  0H;hd2[ 55]:=  0H;hd2[ 56]:= 40H;hd2[ 57]:=  0H;hd2[ 58]:=  2H;hd2[ 59]:=  0H;
  hd2[ 60]:=  0H;hd2[ 61]:=  0H;hd2[ 62]:= 1FH;hd2[ 63]:=  0H;hd2[ 64]:=  2H;hd2[ 65]:=  0H;hd2[ 66]:= 15H;hd2[ 67]:=  0H;hd2[ 68]:= 1EH;hd2[ 69]:=  0H;
  hd2[ 70]:=  8H;hd2[ 71]:=  0H;hd2[ 72]:=  7H;hd2[ 73]:= 47H;hd2[ 74]:= 65H;hd2[ 75]:= 6EH;hd2[ 76]:= 65H;hd2[ 77]:= 72H;hd2[ 78]:= 61H;hd2[ 79]:= 6CH;
  hd2[ 80]:= 1EH;hd2[ 81]:=  0H;hd2[ 82]:=  2H;hd2[ 83]:=  0H;hd2[ 84]:=  1H;hd2[ 85]:= 30H;hd2[ 86]:= 1EH;hd2[ 87]:=  0H;hd2[ 88]:=  5H;hd2[ 89]:=  0H;
  hd2[ 90]:=  4H;hd2[ 91]:= 30H;hd2[ 92]:= 2EH;hd2[ 93]:= 30H;hd2[ 94]:= 30H;hd2[ 95]:= 1EH;hd2[ 96]:=  0H;hd2[ 97]:=  6H;hd2[ 98]:=  0H;hd2[ 99]:=  5H;
  hd2[100]:= 23H;hd2[101]:= 2CH;hd2[102]:= 23H;hd2[103]:= 23H;hd2[104]:= 30H;hd2[105]:= 1EH;hd2[106]:=  0H;hd2[107]:=  9H;hd2[108]:=  0H;hd2[109]:=  8H;
  hd2[110]:= 23H;hd2[111]:= 2CH;hd2[112]:= 23H;hd2[113]:= 23H;hd2[114]:= 30H;hd2[115]:= 2EH;hd2[116]:= 30H;hd2[117]:= 30H;hd2[118]:= 1EH;hd2[119]:=  0H;
  hd2[120]:= 18H;hd2[121]:=  0H;hd2[122]:= 17H;hd2[123]:= 22H;hd2[124]:= 24H;hd2[125]:= 22H;hd2[126]:= 23H;hd2[127]:= 2CH;hd2[128]:= 23H;hd2[129]:= 23H;
  hd2[130]:= 30H;hd2[131]:= 5FH;hd2[132]:= 29H;hd2[133]:= 3BH;hd2[134]:= 5CH;hd2[135]:= 28H;hd2[136]:= 22H;hd2[137]:= 24H;hd2[138]:= 22H;hd2[139]:= 23H;
  hd2[140]:= 2CH;hd2[141]:= 23H;hd2[142]:= 23H;hd2[143]:= 30H;hd2[144]:= 5CH;hd2[145]:= 29H;hd2[146]:= 1EH;hd2[147]:=  0H;hd2[148]:= 1DH;hd2[149]:=  0H;
  hd2[150]:= 1CH;hd2[151]:= 22H;hd2[152]:= 24H;hd2[153]:= 22H;hd2[154]:= 23H;hd2[155]:= 2CH;hd2[156]:= 23H;hd2[157]:= 23H;hd2[158]:= 30H;hd2[159]:= 5FH;
  hd2[160]:= 29H;hd2[161]:= 3BH;hd2[162]:= 5BH;hd2[163]:= 52H;hd2[164]:= 65H;hd2[165]:= 64H;hd2[166]:= 5DH;hd2[167]:= 5CH;hd2[168]:= 28H;hd2[169]:= 22H;
  hd2[170]:= 24H;hd2[171]:= 22H;hd2[172]:= 23H;hd2[173]:= 2CH;hd2[174]:= 23H;hd2[175]:= 23H;hd2[176]:= 30H;hd2[177]:= 5CH;hd2[178]:= 29H;hd2[179]:= 1EH;
  hd2[180]:=  0H;hd2[181]:= 1EH;hd2[182]:=  0H;hd2[183]:= 1DH;hd2[184]:= 22H;hd2[185]:= 24H;hd2[186]:= 22H;hd2[187]:= 23H;hd2[188]:= 2CH;hd2[189]:= 23H;
  hd2[190]:= 23H;hd2[191]:= 30H;hd2[192]:= 2EH;hd2[193]:= 30H;hd2[194]:= 30H;hd2[195]:= 5FH;hd2[196]:= 29H;hd2[197]:= 3BH;hd2[198]:= 5CH;hd2[199]:= 28H;
  hd2[200]:= 22H;hd2[201]:= 24H;hd2[202]:= 22H;hd2[203]:= 23H;hd2[204]:= 2CH;hd2[205]:= 23H;hd2[206]:= 23H;hd2[207]:= 30H;hd2[208]:= 2EH;hd2[209]:= 30H;
  hd2[210]:= 30H;hd2[211]:= 5CH;hd2[212]:= 29H;hd2[213]:= 1EH;hd2[214]:=  0H;hd2[215]:= 23H;hd2[216]:=  0H;hd2[217]:= 22H;hd2[218]:= 22H;hd2[219]:= 24H;
  hd2[220]:= 22H;hd2[221]:= 23H;hd2[222]:= 2CH;hd2[223]:= 23H;hd2[224]:= 23H;hd2[225]:= 30H;hd2[226]:= 2EH;hd2[227]:= 30H;hd2[228]:= 30H;hd2[229]:= 5FH;
  hd2[230]:= 29H;hd2[231]:= 3BH;hd2[232]:= 5BH;hd2[233]:= 52H;hd2[234]:= 65H;hd2[235]:= 64H;hd2[236]:= 5DH;hd2[237]:= 5CH;hd2[238]:= 28H;hd2[239]:= 22H;
  hd2[240]:= 24H;hd2[241]:= 22H;hd2[242]:= 23H;hd2[243]:= 2CH;hd2[244]:= 23H;hd2[245]:= 23H;hd2[246]:= 30H;hd2[247]:= 2EH;hd2[248]:= 30H;hd2[249]:= 30H;
  hd2[250]:= 5CH;hd2[251]:= 29H;hd2[252]:= 1EH;hd2[253]:=  0H;hd2[254]:=  3H;hd2[255]:=  0H;hd2[256]:=  2H;hd2[257]:= 30H;hd2[258]:= 25H;hd2[259]:= 1EH;
  hd2[260]:=  0H;hd2[261]:=  6H;hd2[262]:=  0H;hd2[263]:=  5H;hd2[264]:= 30H;hd2[265]:= 2EH;hd2[266]:= 30H;hd2[267]:= 30H;hd2[268]:= 25H;hd2[269]:= 1EH;
  hd2[270]:=  0H;hd2[271]:=  9H;hd2[272]:=  0H;hd2[273]:=  8H;hd2[274]:= 30H;hd2[275]:= 2EH;hd2[276]:= 30H;hd2[277]:= 30H;hd2[278]:= 45H;hd2[279]:= 2BH;
  hd2[280]:= 30H;hd2[281]:= 30H;hd2[282]:= 1EH;hd2[283]:=  0H;hd2[284]:=  7H;hd2[285]:=  0H;hd2[286]:=  6H;hd2[287]:= 6DH;hd2[288]:= 2FH;hd2[289]:= 64H;
  hd2[290]:= 2FH;hd2[291]:= 79H;hd2[292]:= 79H;hd2[293]:= 1EH;hd2[294]:=  0H;hd2[295]:= 0BH;hd2[296]:=  0H;hd2[297]:= 0AH;hd2[298]:= 64H;hd2[299]:= 5CH;
  hd2[300]:= 2DH;hd2[301]:= 6DH;hd2[302]:= 6DH;hd2[303]:= 6DH;hd2[304]:= 5CH;hd2[305]:= 2DH;hd2[306]:= 79H;hd2[307]:= 79H;hd2[308]:= 1EH;hd2[309]:=  0H;
  hd2[310]:=  7H;hd2[311]:=  0H;hd2[312]:=  6H;hd2[313]:= 64H;hd2[314]:= 5CH;hd2[315]:= 2DH;hd2[316]:= 6DH;hd2[317]:= 6DH;hd2[318]:= 6DH;hd2[319]:= 1EH;
  hd2[320]:=  0H;hd2[321]:=  8H;hd2[322]:=  0H;hd2[323]:=  7H;hd2[324]:= 6DH;hd2[325]:= 6DH;hd2[326]:= 6DH;hd2[327]:= 5CH;hd2[328]:= 2DH;hd2[329]:= 79H;
  hd2[330]:= 79H;hd2[331]:= 1EH;hd2[332]:=  0H;hd2[333]:= 0CH;hd2[334]:=  0H;hd2[335]:= 0BH;hd2[336]:= 68H;hd2[337]:= 3AH;hd2[338]:= 6DH;hd2[339]:= 6DH;
  hd2[340]:= 5CH;hd2[341]:= 20H;hd2[342]:= 41H;hd2[343]:= 4DH;hd2[344]:= 2FH;hd2[345]:= 50H;hd2[346]:= 4DH;hd2[347]:= 1EH;hd2[348]:=  0H;hd2[349]:= 0FH;
  hd2[350]:=  0H;hd2[351]:= 0EH;hd2[352]:= 68H;hd2[353]:= 3AH;hd2[354]:= 6DH;hd2[355]:= 6DH;hd2[356]:= 3AH;hd2[357]:= 73H;hd2[358]:= 73H;hd2[359]:= 5CH;
  hd2[360]:= 20H;hd2[361]:= 41H;hd2[362]:= 4DH;hd2[363]:= 2FH;hd2[364]:= 50H;hd2[365]:= 4DH;hd2[366]:= 1EH;hd2[367]:=  0H;hd2[368]:=  5H;hd2[369]:=  0H;
  hd2[370]:=  4H;hd2[371]:= 68H;hd2[372]:= 3AH;hd2[373]:= 6DH;hd2[374]:= 6DH;hd2[375]:= 1EH;hd2[376]:=  0H;hd2[377]:=  8H;hd2[378]:=  0H;hd2[379]:=  7H;
  hd2[380]:= 68H;hd2[381]:= 3AH;hd2[382]:= 6DH;hd2[383]:= 6DH;hd2[384]:= 3AH;hd2[385]:= 73H;hd2[386]:= 73H;hd2[387]:= 1EH;hd2[388]:=  0H;hd2[389]:= 0DH;
  hd2[390]:=  0H;hd2[391]:= 0CH;hd2[392]:= 6DH;hd2[393]:= 2FH;hd2[394]:= 64H;hd2[395]:= 2FH;hd2[396]:= 79H;hd2[397]:= 79H;hd2[398]:= 5CH;hd2[399]:= 20H;
  hd2[400]:= 68H;hd2[401]:= 3AH;hd2[402]:= 6DH;hd2[403]:= 6DH;

  fmt[  0]:= 1EH;fmt[  1]:=  0H;fmt[  2]:=  9H;fmt[  3]:=  0H;fmt[  4]:=  8H;fmt[  5]:= 23H;fmt[  6]:= 23H;fmt[  7]:= 30H;fmt[  8]:= 2EH;fmt[  9]:= 30H;
  fmt[ 10]:= 45H;fmt[ 11]:= 2BH;fmt[ 12]:= 30H;fmt[ 13]:= 1EH;fmt[ 14]:=  0H;fmt[ 15]:=  6H;fmt[ 16]:=  0H;fmt[ 17]:=  5H;fmt[ 18]:= 6DH;fmt[ 19]:= 6DH;
  fmt[ 20]:= 3AH;fmt[ 21]:= 73H;fmt[ 22]:= 73H;fmt[ 23]:= 1EH;fmt[ 24]:=  0H;fmt[ 25]:=  2H;fmt[ 26]:=  0H;fmt[ 27]:=  1H;fmt[ 28]:= 40H;

  wpr[  0]:= 19H;wpr[  1]:=  0H;wpr[  2]:=  2H;wpr[  3]:=  0H;wpr[  4]:=  0H;wpr[  5]:=  0H;
  pro[  0]:= 12H;pro[  1]:=  0H;pro[  2]:=  2H;pro[  3]:=  0H;pro[  4]:=  1H;pro[  5]:=  0H;
  xfr[  0]:= 43H;xfr[  1]:=  0H;xfr[  2]:=  4H;xfr[  3]:=  0H;xfr[  4]:=  0H;xfr[  5]:=  0H;xfr[  6]:= 40H;xfr[  7]:=  0H;
  dcw[  0]:= 55H;dcw[  1]:=  0H;dcw[  2]:=  2H;dcw[  3]:=  0H;dcw[  4]:= 0AH;dcw[  5]:=  0H;

  dim[  0]:=  0H;dim[  1]:=  0H;dim[  2]:=  8H;dim[  3]:=  0H;dim[  4]:=  0H;dim[  5]:=  0H;dim[  6]:=  0H;dim[  7]:=  0H;dim[  8]:=  0H;dim[  9]:=  0H;
  dim[ 10]:=  0H;dim[ 11]:=  0H;

  tail[ 0]:=3DH;tail[ 1]:= 0H;tail[ 2]:=0AH;tail[ 3]:= 0H;tail[ 4]:=78H;tail[ 5]:= 0H;tail[ 6]:=3CH;tail[ 7]:= 0H;tail[ 8]:=4CH;tail[ 9]:=3BH;
  tail[10]:=81H;tail[11]:=24H;tail[12]:= 0H;tail[13]:= 0H;tail[14]:=3EH;tail[15]:= 0H;tail[16]:=0EH;tail[17]:= 0H;tail[18]:= 0H;tail[19]:= 1H;
  tail[20]:= 1H;tail[21]:= 0H;tail[22]:= 1H;tail[23]:= 0H;tail[24]:= 0H;tail[25]:= 0H;tail[26]:= 0H;tail[27]:= 1H;tail[28]:= 0H;tail[29]:= 0H;
  tail[30]:= 0H;tail[31]:= 0H;tail[32]:=1DH;tail[33]:= 0H;tail[34]:=0FH;tail[35]:= 0H;tail[36]:= 3H;tail[37]:= 0H;tail[38]:= 0H;tail[39]:= 0H;
  tail[40]:= 0H;tail[41]:= 0H;tail[42]:= 0H;tail[43]:= 1H;tail[44]:= 0H;tail[45]:= 0H;tail[46]:= 0H;tail[47]:= 0H;tail[48]:= 0H;tail[49]:= 0H;
  tail[50]:= 0H;tail[51]:=0AH;tail[52]:= 0H;tail[53]:= 0H;tail[54]:= 0H;
END xls.
